// Mac Pierce 2021
/// Code to run the project 'A Scanner Darkly'
// www.macpierce.com/onbeingseen

// 1 of 2 code blocks, TX - Transmit, uploaded to the conductor
// Made for the M5-Stack Core ESP32 Module

/* **** Update Log *******
 *  - 210410 - v0-01
 *    - Building it up, settled on painless mesh for mesh network
 *  - 210427 - v0-02
 *    - Starting to build for the M5 Stack Basic ESP32 Module.
 *    - May have a conflict with the painless mesh library and the M5 Library...
 *  - 210508 - v0-03
 *    - Continuing...
 *    - Needed a 'Callback method prototype' or two.
 *    - Incorporating dot sequence function - New vals included - Tested, works.
 *    - Built UI, buttons and LCD now mostly working, need to test.
 *    - UI Is good. LCD now automatically goes to sleep after 10 seconds.
 *  - 210514 - v0-04
 *    - Built in debug modes (3) with UI control (Hold button C for 5 seconds to toggle debug)
 *    - Need to troubleshoot why screen immediately goes to sleep after debug mode is turned off.
 *  - 210706 - v0-05
 *    - Built in SD Card read functionality, reads text.txt and writes current position to log.txt
 *  - 210707 - v2-0
 *    - Attempting to incorporate ESP-Now for wifi
 *    - https://www.youtube.com/watch?v=cIKX19esGQU (?)
 *    - Leaving off on line 191, ESP32 module freezes when current code uploaded.
 *  - 210710 - v2-01
 *    - Appears that task scheduler and ESP-Now are incompatible. Removing Task Scheduler. 
 *    - Removed Task Scheduler, need to build out timing system.
 *    - Built timing system, need to test.
 *    - Built Error State mode.
 *  - 210714 - v3-0
 *    - Removed Painless Mesh
 *    - Incoporated ESP-Now
 *    - DONE !
*/

// ---------- Used Libraries --------------------------------------------------------------------

#include <esp_now.h>
#include <WiFi.h>
#include <M5Stack.h>

#define MY_NAME         "OBS_CONDUCTOR"
#define MY_ROLE         ESP_NOW_ROLE_CONTROLLER
#define RECEIVER_ROLE   ESP_NOW_ROLE_SLAVE
#define WIFI_CHANNEL    1

// ----------- Variable Def ---------------------------------------------------------------------

// --- Project variables ---
String charPattern = "000000000000000"; // Holds the pattern for the most reacently read character.
String charSpace = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789 /:;=+{}()[].,?!*@#$%-_|^><"; // The character space that the system can display
char cChar = '\0'; // The current character loaded to be displayed, set to a null char for initializeation;

// --- State booleans ---
bool screenLocked = true; // Whether the UI is locked
bool enabled = true; // Whether the whole mechanism is enabled and sending content
bool paused = false; // Whether the system is paused on a letter
bool lcdOn = true; // Whether the LCD should be enabled
bool btna = false; // State holder of button A
bool btnb = false; // State holder of button B
bool btnc = false; // State holder of button C

// --- Timing ---
const int txInt = 1000; // 1 second interval at which a signal is transmitted from the conductor.
const int lcdDisableInt = 10000; // 10 second timeout for the LCD to turn off after the UI is not used.
const int bntChkInt = 100; // 100 millisecond interval that checks for a change state in the UI buttons.
const int dbgCheckInt = 3000; // 3 second interval countdown to see if the debug mode is toggled.
float cMil; // Container for the current millisecond read at the top of the main loop.
float pTXMil; // Last time the SD read was triggered.
float pLCDMil; // Last time the LCD Shutdown was triggered.
float pBTNMil; // Last time the Button read was triggered.
float pDBGMil; // Last time the Debug toggle was triggered.

// --- SD ---
File textFile; // Container for the text file.
File logFile; // Container for the log file.
long txtSize; // Shows how large the text.txt file on the SD Card is.
long cRead; // The current position of the system in reading off the contents of the text.txt.

// --- WiFi-Now ---
uint8_t broadcastAddress[] = {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF}; // Sets the send address of all communications to broadcast.

// --- Debug --- 
bool debug = true; // ******** ENABLED FOR TESTING **************
bool error = false; // Flag for an error state
bool dbgBtnPress; // Checks to see how long the debug button is pressed for.
const int dbgModes = 3; // How many debug modes there are
int debugMode = 0;
int testCount = 0;

// --------------------- Project Setup ---------------------------------------------------------------
void setup(){

  M5.begin(); // Starts the LCD
  M5.Power.begin(); // **** What is this? ****

  Serial.begin(115200);
  if(Serial){
    Serial.println(' ');
    Serial.println("A SCANNER DARKLY");
    Serial.println("Mac Pierce - 2021");
    Serial.println(' ');
  }


  // ------------ ESP-Now Stuff------------
  WiFi.mode(WIFI_STA);
  if(Serial){
    Serial.println();
    Serial.println("Conductor MAC Address - " + String(WiFi.macAddress())); 
  }
  if(esp_now_init() != ESP_OK){
    Serial.println("CRITICAL ERROR - FAILURE INITIALIZING ESP-NOW, Contact Mac Pierce");
    return;
  }
  esp_now_peer_info_t peerInfo;
  //esp_now_set_self_role(ESP_NOW_ROLE_CONTROLLER); *********************
  memcpy(peerInfo.peer_addr,broadcastAddress,6);
  peerInfo.channel = 0;
  peerInfo.encrypt = false;

  if(esp_now_add_peer(&peerInfo) != ESP_OK){
    Serial.println("Failed to add peer");
    //return; ***************************************
  }

  // ------------ LCD Stuff ---------------
  M5.Lcd.setBrightness(255);
  M5.Lcd.setTextDatum(MC_DATUM);
  M5.Lcd.fillScreen(BLACK);

  // ------------ SD Stuff ----------------
  if(SD.begin()){
    textFile = SD.open("/text.txt",FILE_READ); // Where the text is read from --------------------------------
    logFile = SD.open("/log.txt",FILE_READ); // Where the current position of reading is held --------------
    if(textFile){
      txtSize = textFile.size();
    }
    else{
      Serial.println("ERROR - TEXT FILE FAILED TO OPEN, text file to be read off of SD Card MUST be named text.txt");
      error = true;
    }
    if(logFile){ // If the log file is available, it 
        logFile.seek(0);
        String pRead = logFile.readStringUntil('\r');
        cRead = pRead.toInt();
        if(cRead >= txtSize){ // Checks to see if it's at the end of the file, resets to start it if so.
            textFile.seek(0);
        }  
        else{
          textFile.seek(cRead);
        }
        logFile.close();
    }
    else{
      Serial.println("ERROR - LOG FILE FAILED TO OPEN");
      error = true;
    }
  }
  else{ // Throws an error state if the SD Card does not initialize
    Serial.println("ERROR - SD FAILED TO INITIALIZE, make sure it is fully seated");
    error = true;
  }
  
  lcdUpdate(true);
  
  if(!error){
    if(Serial){
      Serial.println("SD Card read success - starting at position " + String(cRead));
    }
  }
  else{ // If the system is in a error state, then it stops everything
    //*********** Build LCD Message
    if(Serial){
      Serial.println("System is in an error mode, please resolve the error and restart the conductor");
    }
    return;
  }
}

// ----------------------- Main Loop ---------------------------------------------------------------------

void loop() {
  cMil = millis();
  if(cMil >= (bntChkInt + pBTNMil)){ // Checks the status of the buttons and re-sends the charater sequence every 100ms
    btnCheck();
    broadcastMSG();
    pBTNMil = cMil;
  }
  if((cMil >= (dbgCheckInt + pDBGMil))&& dbgBtnPress){ // Checks to see if it's been 3 seconds that the debug button has been pressed
    debugToggle();
    dbgBtnPress = false;
  }
  if((cMil >= (txInt + pTXMil)) && (!paused && enabled)){ // Checks to see if its been more than a second since the last message was send, and if the mechanism is enabled and not paused
    sendCharSet(); 
    pTXMil = cMil;
  }
  if((cMil >= (lcdDisableInt + pLCDMil)) && !debug){ // If debug mode is off and its been greater than 10 seconds since the last button was pressed, it shuts off the LCD.
    lcdSleep();
  }
}

// ----------------------- Main Project Functions ------------------------------------------------------------

void sendCharSet() { // Reads the SD Card, assigns the dot matrix series and sends out the char binary sequence once per interval.
  if(debug){ // If the debug mode is enabled, it sends along one of the debug modes
    debugTests();
  }
  else{ // What runs for the install - Reads off each letter of the text and sends
    if(!logFile){ // Opens the log file
      logFile = SD.open("/log.txt",FILE_READ);
      if(!logFile){
        Serial.println("ERROR - LOG FILE FAILED TO OPEN, reverting to read-only mode");
        error = true;
      }
    }
    if(textFile.available()){ // Reads off the char at the position indicated by the log file.
        cChar = textFile.read();
        cRead = textFile.position();
        logFlush(cRead);
        if(cRead >= txtSize){ //Checks to see if it's at the end of the file, resets to start it if so.
          textFile.seek(0);
        }
      }
    else{
        Serial.println("ERROR - TEXT FILE FAILED TO OPEN, reverting to debug mode");
        error = true;
    }
    charAssign(cChar);
  }

  //Serial.println("Test"); //*******************************************************************************
  broadcastMSG(); // function to handle all the ESP-Now sending
  
  if(lcdOn){
    lcdUpdate(false); // Updates the LCD with the current charachter
  }
  if(Serial){
    Serial.println("Showing : " + String(cChar) + " - " + charPattern);
  }
}

// ----------------------- Char assignment code --------------------------------------------------------

void charAssign(char character){
  if(isAlphaNumeric(character)){ // Converts all lowercase letters to uppercase
    if(isLowerCase(character)){
      character -= ('a'-'A');
    }
  }
  String dotMatrix;
  switch (character){
    case '\0': //----------- Sending a Null char will set it all on ------------
      dotMatrix = "111111111111111";
      break;
    case 'A': //---------------- The boolean sequence of each character segment --------------------
      dotMatrix="111101111101101"; 
      break;
    case 'B':
      dotMatrix="110101110101110";
      break;
    case 'C':
      dotMatrix="111100100100111";
      break;
    case 'D':
      dotMatrix="110101101101110";
      break;
    case 'E':
      dotMatrix="111100111100111";
      break;
    case 'F':
      dotMatrix="111100111100100";
      break;
    case 'G':
      dotMatrix="011100111101010";
      break;
    case 'H':
      dotMatrix="101101111101101";
      break;
    case 'I':
      dotMatrix="111010010010111";
      break;
    case 'J':
      dotMatrix="001001001101010";
      break;
    case 'K':
      dotMatrix="101101110101101";
      break;
    case 'L':
      dotMatrix="100100100100111";
      break;
    case 'M':
      dotMatrix="101111101101101";
      break;
    case 'N':
      dotMatrix="110101101101101";
      break;
    case 'O':
      dotMatrix="010101101101010";
      break;
    case 'P':
      dotMatrix="111101111100100";
      break;
    case 'Q':
      dotMatrix="111101101111001";
      break;
    case 'R':
      dotMatrix="111101110101101";
      break;
    case 'S':
      dotMatrix="111100111001111";
      break;
    case 'T':
      dotMatrix="111010010010010";
      break;
    case 'U':
      dotMatrix="101101101101111";
      break;
    case 'V':
      dotMatrix="101101101101010";
      break;
    case 'W':
      dotMatrix="101101101111101";
      break;
    case 'X':
      dotMatrix="101101010101101";
      break;
    case 'Y':
      dotMatrix="101101010010010";
      break;
    case 'Z':
      dotMatrix="111001010100111";
      break;
    case '0':
      dotMatrix="111101101101111";
      break;
    case '1':
      dotMatrix="110010010010111";
      break;
    case '2':
      dotMatrix="111001111100111";
      break;
    case '3':
      dotMatrix="111001111001111";
      break;
    case '4':
      dotMatrix="101101111001001";
      break;
    case '5':
      dotMatrix="111100111001111";
      break;
    case '6':
      dotMatrix="111100111101111";
      break;
    case '7':
      dotMatrix="111001001001001";
      break;
    case '8':
      dotMatrix="111101111101111";
      break;
    case '9':
      dotMatrix="111101111001001";
      break;
    case '/':
      dotMatrix="001001010100100";
      break;
    case ',':
      dotMatrix="000000000010100";
      break;
    case ':':
      dotMatrix="000010000010000";
      break;
    case ';':
      dotMatrix="000010000010100";
      break;
    case '=':
      dotMatrix="000111000111000";
      break;
    case '+':
      dotMatrix="000010111010000";
      break;
    case '{':
      dotMatrix="001010100010001";
      break;
    case '}':
      dotMatrix="100010001010100";
      break;
    case '(':
      dotMatrix="001010010010001";
      break;
    case ')':
      dotMatrix="100010010010100";
      break;
    case '[':
      dotMatrix="011010010010011";
      break;
    case ']':
      dotMatrix="110010010010110";
      break;
    case '.':
      dotMatrix="000000000000010";
      break;
    case '?':
      dotMatrix="110001010000010";
      break;
    case '!':
      dotMatrix="010010010000010";
      break;
    case '*':
      dotMatrix="000111111111000";
      break;
    case '@':
      dotMatrix="111001111101111";
      break;
    case '#':
      dotMatrix="101111101111101";
      break;
    case '$':
      dotMatrix="111110111011111";
      break;
    case '%':
      dotMatrix="101001010100101";
      break;
    case '-':
      dotMatrix="000000110000000";
      break;
    case '_':
      dotMatrix="000000111000000";
      break;
    case '|':
      dotMatrix="010010010010010";
      break;
    case '^':
      dotMatrix="010101000000000";
      break;
    case '>':
      dotMatrix="100010001010100";
      break;
    case '<':
      dotMatrix="001010100010001";
      break;
    default: //----------- Will default to an empty character if no useable chars are sent ----------
      dotMatrix="000000000000000";
      break;
  }
  charPattern = dotMatrix;
}

// ---------------------------------------------------- UI Elements, Buttons & LCD ---------------------------------------------

void btnCheck() { // Checks the state of the buttons, triggers UI elements
  bool trigA = false ; bool trigB = false; bool trigC = false;

  // -----------------------------------These read the state of each of the buttons
  if(M5.BtnA.read()){ // Reads button A
    btna = true;
  }
  else if(btna){
    trigA = true;
    btna = false;
    if(debug&&Serial){
      Serial.print("Button A Pressed - ");
    }
  }
  if(M5.BtnB.read()){ // Reads button B
    btnb = true;
  }
  else if(btnb){
    trigB = true;
    btnb = false;
    if(debug&&Serial){
      Serial.print("Button B Pressed - ");
    }
  }
  if(M5.BtnC.read()){ // Reads button C
    btnc = true;
    if(!screenLocked){
      if(!dbgBtnPress){
        pDBGMil = millis(); // If the debug button hasnt been pressed previously;
        dbgBtnPress = true;
      }
    }
  }
  else if(btnc){
    trigC = true;
    btnc = false;
    dbgBtnPress = false;
    if(debug&&Serial){
      Serial.print("Button C Pressed - ");
    }
  }

  if(trigA||trigB||trigC){ 
    if(lcdOn){ // Only enabled these if the LCD is on
      if(trigA){ // ---------------------- Button 1 is a UI lock
        screenLock();
      }
      else if(trigB){ // ----------------- Button 2 is a the pause function
        if(!screenLocked && enabled){
          functPause();
        }
      }
      else if(trigC){ // ----------------- Button 3 is the disable all function (Sleep mode);
        if(!screenLocked){
          if(!debug){
            functStop();
          }
          else{
            debugMode ++;
            testCount = 0;
          }
        }
      }
    }
    else{ // If any of the buttons are pressed, it wakes up the LCD
      lcdWake();
    }
    lcdUpdate(true);
    pLCDMil = millis();
  }
}

void lcdUpdate(bool fullRef){ // ----------------- The function that updated the LCD on the M5 -----------------
  if(fullRef){ // The header info, stuff that doesnt change
    M5.Lcd.clear(TFT_BLACK);
    M5.Lcd.setTextColor(TFT_WHITE, TFT_BLACK);
    M5.Lcd.setFreeFont(&FreeMonoBold12pt7b);
    M5.Lcd.drawString("ON BEING SEEN", 67, 10); // Shows the name of the project
    M5.Lcd.setFreeFont(&FreeMonoBold9pt7b);
    M5.Lcd.drawString("Mac Pierce - 2021", 67, 30); // Artists name and date (ME!)
    if(screenLocked){
      M5.Lcd.setTextColor(TFT_DARKGREY, TFT_BLACK);
      M5.Lcd.drawString("UNLOCK", 67, 225); // Text for lock button
    }
    else{
      M5.Lcd.setTextColor(TFT_LIGHTGREY, TFT_BLACK);
      M5.Lcd.drawString("LOCK", 67, 225); // Text for lock button
      
      if(enabled){
        if(debug){
          M5.Lcd.setTextColor(TFT_RED, TFT_BLACK);
          M5.Lcd.drawString("DEBUG", 255, 225); // Text for stop button - stop
        }
        else{
          M5.Lcd.setTextColor(TFT_RED, TFT_BLACK);
          M5.Lcd.drawString("STOP", 255, 225); // Text for stop button - stop
        }
        if(paused){
          M5.Lcd.setTextColor(TFT_GREEN, TFT_BLACK);
          M5.Lcd.drawString("RESUME", 160, 225); // Text for pause button - resume
        }
        else{
          M5.Lcd.setTextColor(TFT_YELLOW, TFT_BLACK);
          M5.Lcd.drawString("PAUSE", 160, 225); // Text for pause button - pause
        }
      }
      else{
        M5.Lcd.setTextColor(TFT_GREEN, TFT_BLACK);
        M5.Lcd.drawString("START", 255, 225); // Text for stop button - start
      }
    }
  }
  
  M5.Lcd.fillRect(0,60,320,160,TFT_BLACK); // Blacks out the center area of the screen with content, flashes the screen less than clear.
  if(enabled){
    M5.Lcd.drawRect(130,82,62,65,TFT_DARKGREY); // Box for current char
    M5.Lcd.setTextColor(TFT_WHITE, TFT_BLACK);
    M5.Lcd.setFreeFont(&FreeMonoBold24pt7b);
    M5.Lcd.drawString(String(cChar), 160, 110); // Shows the current character
    M5.Lcd.setTextColor(TFT_LIGHTGREY, TFT_BLACK);
    M5.Lcd.setFreeFont(&FreeMonoBold12pt7b);
    M5.Lcd.drawString(charPattern, 160, 170); // Shows the current dotmatrix series
    M5.Lcd.setTextColor(TFT_WHITE, TFT_BLACK);
    M5.Lcd.setFreeFont(&FreeMonoBold24pt7b);
    if(debug){
      if(debugMode == 0){
        if(cChar == '\0'){
          M5.Lcd.fillRect(130,82,62,65,TFT_WHITE); // Box for current char
        }
      }
      else if(debugMode == 1){
        M5.Lcd.drawString(String(cChar), 160, 110); // Shows the current character
      }
      else if(debugMode == 2){
        M5.Lcd.drawString(String(testCount), 160, 110); // Shows the count of the test
      }
      M5.Lcd.setTextColor(TFT_RED, TFT_BLACK);
      M5.Lcd.setFreeFont(&FreeMonoBold12pt7b);
      M5.Lcd.drawString("DEBUG MODE "+String(debugMode+1), 160, 60); // Writes the debug mode
    }
    else{
      M5.Lcd.drawString(String(cChar), 160, 110); // Shows the current character
    }
  }
  else{
    M5.Lcd.setTextColor(TFT_WHITE, TFT_BLACK);
    M5.Lcd.setFreeFont(&FreeMonoBold24pt7b);
    M5.Lcd.drawString("OUTPUT OFF", 160, 120); // Shows the current character
  }
  if (!textFile) {
    M5.Lcd.setTextColor(TFT_RED, TFT_BLACK);
    M5.Lcd.setFreeFont(&FreeMonoBold12pt7b);
    M5.Lcd.drawString("CARD FAILURE", 160, 195); // Shows the current dotmatrix series
  }
  else{
    M5.Lcd.setTextColor(TFT_DARKGREY, TFT_BLACK);
    M5.Lcd.setFreeFont(&FreeMonoBold9pt7b);
    M5.Lcd.drawString("Reading: " + String(cRead) + " of " + String(txtSize), 160, 195); // Shows the current dotmatrix series
  }
}

void broadcastMSG(){ // Sends the messages and converts from a string to bytes to send. (https://stackoverflow.com/questions/64998264/cannot-send-string-properly-with-esp-now)
  uint8_t *buffer = (uint8_t*) charPattern.c_str();
  size_t sizeBuff = sizeof(buffer) * charPattern.length();
  esp_err_t result = esp_now_send(broadcastAddress, buffer, sizeBuff);

  if(result == ESP_OK){
    if(Serial){
      Serial.println("Pattern broadcast with success");
      Serial.println();
    }
  }
  else{
    if(Serial){
      Serial.println("Error sending the data");
    }
    //error = true; *******************
  }
}

void debugTests(){ // What performs all of the Debug functions when debug mode is enabled
  switch(debugMode){
      case 0: // ------------------- Default flashes all on & off -------------------
        if((testCount%2) >= 1){
          cChar = '\0';
        }
        else{
          cChar = ' ';
        }
        charAssign(cChar);
        
        break;
    case 1: // ------------------- Reads off all letters it can show -------------------
      if(testCount >= charSpace.length()-1){
        testCount = 0;
      }
      cChar = charSpace.charAt(testCount);
      charAssign(cChar);
      break;
    case 2: // ------------------- Each node turns on in series, bypasses charAssign -------------------
      if(testCount > charPattern.length()-1){
        testCount = 0;
      }
      cChar = ' ';
      for(int x = 0; x <= charPattern.length()-1; x++){
        if(x == testCount){
          charPattern.setCharAt(x,'1');
        }
        else{
          charPattern.setCharAt(x,'0');
        }
      }
      break;
    default: // --------- If the mode # is higher than the number of nodes, it returns it to 0 ---------
      debugMode = 0;
      break;
  }
  testCount++;
}

void lcdSleep(){ // Puts the LCD to sleep
  lcdOn = false;
  screenLocked = true;
  M5.Lcd.sleep();
  M5.Lcd.setBrightness(0);
  if(debug&&Serial){
    Serial.println("Screen going to sleep...");
  }
}

void lcdWake(){ // Wakes up the LCD Screen
  lcdOn = true;
  M5.Lcd.wakeup();
  M5.Lcd.setBrightness(255);
  if(debug&&Serial){
    Serial.println("Screen Resuming functionality");
  }
}

void screenLock(){ // Toggles the screen lock
  screenLocked = !screenLocked;
  if(debug&&Serial){
      Serial.println("Screen lock toggled.");
  }
}

void functPause(){ // Toggles the functionality of the system, based off of the UI inputs - PAUSE
  paused = !paused;
  if(paused){
    broadcastMSG();
    if(debug&&Serial){
      Serial.println("Playback Paused.");
    }
  }
  else{
    if(debug&&Serial){
      Serial.println("Playback Resumed.");
    }
  }
}

void functStop(){ // Toggles the functionality of the system, based off of the UI inputs - STOP
  enabled = !enabled;
  if(enabled){
    if(debug&&Serial){
      Serial.println("Blackout ended - Playback Started.");
    }
  }
  else{
    charAssign(' '); // Sends a space char to set all the spotlights off
    broadcastMSG();
    paused = false;
    if(debug&&Serial){
      Serial.println("Full Stop - Blackout.");
    }
  }
}

void debugToggle(){
  debug = !debug;
  if(debug && Serial){
    Serial.println("Debug toggled");
  }
}

void logFlush(long x){ // Flushes the current data of the log file and writes the new log file.
  SD.remove("/log.txt");
  logFile = SD.open("/log.txt",FILE_WRITE);
  if(!logFile){
    Serial.println("ERROR - LOG FILE FAILED TO OPEN, reverting to debug mode");
    error = true;
  }
  logFile.println(x);
  logFile.close();
}
